@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2005 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2005 Laszlo Molnar
+   Copyright (C) 1996-2007 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2007 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -183,81 +183,143 @@ inline void set_le64(void *p, acc_uint64l_t v)
 // get signed values, i.e. sign-extend
 **************************************************************************/
 
-inline int sign_extend(int v, int bits)
+inline int sign_extend(unsigned v, unsigned bits)
 {
     const unsigned sign_bit = 1u << (bits - 1);
-    v |= 0u - (v & sign_bit);
-    return v;
+    v &= sign_bit | (sign_bit - 1);
+    //v = (v ^ sign_bit) - sign_bit;
+    v |= 0 - (v & sign_bit);
+    return (int) v;
 }
 
-inline acc_int64l_t sign_extend(acc_int64l_t v, int bits)
+inline acc_int64l_t sign_extend(acc_uint64l_t v, unsigned bits)
 {
     const acc_uint64l_t sign_bit = ACC_UINT64_C(1) << (bits - 1);
-    v |= ACC_UINT64_C(0) - (v & sign_bit);
-    return v;
+    v &= sign_bit | (sign_bit - 1);
+    //v = (v ^ sign_bit) - sign_bit;
+    v |= 0 - (v & sign_bit);
+    return (acc_int64l_t) v;
 }
 
 inline int get_be16_signed(const void *p)
 {
-    int v = get_be16(p);
+    unsigned v = get_be16(p);
     return sign_extend(v, 16);
 }
 
 inline int get_be24_signed(const void *p)
 {
-    int v = get_be24(p);
+    unsigned v = get_be24(p);
     return sign_extend(v, 24);
 }
 
 inline int get_be32_signed(const void *p)
 {
-    int v = get_be32(p);
+    unsigned v = get_be32(p);
     return sign_extend(v, 32);
 }
 
 inline acc_int64l_t get_be64_signed(const void *p)
 {
-    acc_int64l_t v = get_be64(p);
+    acc_uint64l_t v = get_be64(p);
     return sign_extend(v, 64);
 }
 
 inline int get_le16_signed(const void *p)
 {
-    int v = get_le16(p);
+    unsigned v = get_le16(p);
     return sign_extend(v, 16);
 }
 
 inline int get_le24_signed(const void *p)
 {
-    int v = get_le24(p);
+    unsigned v = get_le24(p);
     return sign_extend(v, 24);
 }
 
 inline int get_le32_signed(const void *p)
 {
-    int v = get_le32(p);
+    unsigned v = get_le32(p);
     return sign_extend(v, 32);
 }
 
-inline int get_le64_signed(const void *p)
+inline acc_int64l_t get_le64_signed(const void *p)
 {
-    acc_int64l_t v = get_le64(p);
+    acc_uint64l_t v = get_le64(p);
     return sign_extend(v, 64);
 }
 
 
+/*************************************************************************
+// swab (bswap)
+**************************************************************************/
+
+inline unsigned acc_swab16(unsigned v)
+{
+    return ((v & 0x00ff) << 8) |
+           ((v & 0xff00) >> 8);
+}
+
+inline unsigned acc_swab32(unsigned v)
+{
+    return ((v & 0x000000ff) << 24) |
+           ((v & 0x0000ff00) <<  8) |
+           ((v & 0x00ff0000) >>  8) |
+           ((v & 0xff000000) >> 24);
+}
+
+
+inline unsigned acc_swab16p(const acc_uint16e_t *p)
+{
+    return acc_swab16(*p);
+}
+
+inline unsigned acc_swap32p(const acc_uint32e_t *p)
+{
+    return acc_swab32(*p);
+}
+
+
+inline void acc_swab16s(acc_uint16e_t *p)
+{
+    *p = acc_swab16(*p);
+}
+
+inline void acc_swab32s(acc_uint32e_t *p)
+{
+    *p = acc_swab32(*p);
+}
+
+
+inline void acc_ua_swab16s(void *p)
+{
+    set_be16(p, get_le16(p));
+}
+
+inline void acc_ua_swab32s(void *p)
+{
+    set_be32(p, get_le32(p));
+}
+
+
 /*************************************************************************
 // classes for portable unaligned access
 //
 // Important: these classes must be PODs (Plain Old Data), i.e. no
 //   constructor, no destructor, no virtual functions and no default
 //   assignment operator, and all fields must be public(!).
+//
+// [Actually we _can_ use a safe non-POD subset, but for this we need
+//  to have gcc bug 17519 fixed - see http://gcc.gnu.org/PR17519 ]
 **************************************************************************/
 
 struct BE16
 {
     unsigned char d[2];
 
+    //inline BE16() { }
+    //BE16(unsigned v) { set_be16(d, v); }
+
     BE16& operator =  (unsigned v) { set_be16(d, v); return *this; }
     BE16& operator += (unsigned v) { set_be16(d, get_be16(d) + v); return *this; }
     BE16& operator -= (unsigned v) { set_be16(d, get_be16(d) - v); return *this; }
@@ -278,6 +340,9 @@ struct BE32
 {
     unsigned char d[4];
 
+    //inline BE32() { }
+    //BE32(unsigned v) { set_be32(d, v); }
+
     BE32& operator =  (unsigned v) { set_be32(d, v); return *this; }
     BE32& operator += (unsigned v) { set_be32(d, get_be32(d) + v); return *this; }
     BE32& operator -= (unsigned v) { set_be32(d, get_be32(d) - v); return *this; }
@@ -298,6 +363,9 @@ struct BE64
 {
     unsigned char d[8];
 
+    //inline BE64() { }
+    //BE64(acc_uint64l_t v) { set_be64(d, v); }
+
     BE64& operator =  (acc_uint64l_t v) { set_be64(d, v); return *this; }
     BE64& operator += (acc_uint64l_t v) { set_be64(d, get_be64(d) + v); return *this; }
     BE64& operator -= (acc_uint64l_t v) { set_be64(d, get_be64(d) - v); return *this; }
@@ -318,6 +386,9 @@ struct LE16
 {
     unsigned char d[2];
 
+    //inline LE16() { }
+    //LE16(unsigned v) { set_le16(d, v); }
+
     LE16& operator =  (unsigned v) { set_le16(d, v); return *this; }
     LE16& operator += (unsigned v) { set_le16(d, get_le16(d) + v); return *this; }
     LE16& operator -= (unsigned v) { set_le16(d, get_le16(d) - v); return *this; }
@@ -338,6 +409,9 @@ struct LE32
 {
     unsigned char d[4];
 
+    //inline LE32() { }
+    //LE32(unsigned v) { set_le32(d, v); }
+
     LE32& operator =  (unsigned v) { set_le32(d, v); return *this; }
     LE32& operator += (unsigned v) { set_le32(d, get_le32(d) + v); return *this; }
     LE32& operator -= (unsigned v) { set_le32(d, get_le32(d) - v); return *this; }
@@ -358,6 +432,9 @@ struct LE64
 {
     unsigned char d[8];
 
+    //inline LE64() { }
+    //LE64(acc_uint64l_t v) { set_le64(d, v); }
+
     LE64& operator =  (acc_uint64l_t v) { set_le64(d, v); return *this; }
     LE64& operator += (acc_uint64l_t v) { set_le64(d, get_le64(d) + v); return *this; }
     LE64& operator -= (acc_uint64l_t v) { set_le64(d, get_le64(d) - v); return *this; }
@@ -417,11 +494,47 @@ template <class T> T* operator + (const LE64& v, T* ptr);
 template <class T> T* operator - (T* ptr, const LE64& v);
 
 
+/*************************************************************************
+// global overloads
+**************************************************************************/
+
+#if 1 && !defined(ALIGN_DOWN)
+inline unsigned ALIGN_DOWN(unsigned a, const LE32& b) { return ALIGN_DOWN(a, (unsigned) b); }
+inline unsigned ALIGN_DOWN(const LE32& a, unsigned b) { return ALIGN_DOWN((unsigned) a, b); }
+inline unsigned ALIGN_UP  (unsigned a, const LE32& b) { return ALIGN_UP  (a, (unsigned) b); }
+inline unsigned ALIGN_UP  (const LE32& a, unsigned b) { return ALIGN_UP  ((unsigned) a, b); }
+#endif
+
+
+#if !defined(UPX_MAX)
+inline unsigned UPX_MAX(unsigned a, const BE16& b)    { return UPX_MAX(a, (unsigned) b); }
+inline unsigned UPX_MAX(const BE16& a, unsigned b)    { return UPX_MAX((unsigned) a, b); }
+inline unsigned UPX_MIN(unsigned a, const BE16& b)    { return UPX_MIN(a, (unsigned) b); }
+inline unsigned UPX_MIN(const BE16& a, unsigned b)    { return UPX_MIN((unsigned) a, b); }
+
+inline unsigned UPX_MAX(unsigned a, const BE32& b)    { return UPX_MAX(a, (unsigned) b); }
+inline unsigned UPX_MAX(const BE32& a, unsigned b)    { return UPX_MAX((unsigned) a, b); }
+inline unsigned UPX_MIN(unsigned a, const BE32& b)    { return UPX_MIN(a, (unsigned) b); }
+inline unsigned UPX_MIN(const BE32& a, unsigned b)    { return UPX_MIN((unsigned) a, b); }
+
+inline unsigned UPX_MAX(unsigned a, const LE16& b)    { return UPX_MAX(a, (unsigned) b); }
+inline unsigned UPX_MAX(const LE16& a, unsigned b)    { return UPX_MAX((unsigned) a, b); }
+inline unsigned UPX_MIN(unsigned a, const LE16& b)    { return UPX_MIN(a, (unsigned) b); }
+inline unsigned UPX_MIN(const LE16& a, unsigned b)    { return UPX_MIN((unsigned) a, b); }
+
+inline unsigned UPX_MAX(unsigned a, const LE32& b)    { return UPX_MAX(a, (unsigned) b); }
+inline unsigned UPX_MAX(const LE32& a, unsigned b)    { return UPX_MAX((unsigned) a, b); }
+inline unsigned UPX_MIN(unsigned a, const LE32& b)    { return UPX_MIN(a, (unsigned) b); }
+inline unsigned UPX_MIN(const LE32& a, unsigned b)    { return UPX_MIN((unsigned) a, b); }
+#endif
+
+
 /*************************************************************************
 // misc
 **************************************************************************/
 
 // for use with qsort()
+extern "C" {
 int __acc_cdecl_qsort be16_compare(const void *, const void *);
 int __acc_cdecl_qsort be24_compare(const void *, const void *);
 int __acc_cdecl_qsort be32_compare(const void *, const void *);
@@ -438,6 +551,7 @@ int __acc_cdecl_qsort le16_compare_signed(const void *, const void *);
 int __acc_cdecl_qsort le24_compare_signed(const void *, const void *);
 int __acc_cdecl_qsort le32_compare_signed(const void *, const void *);
 int __acc_cdecl_qsort le64_compare_signed(const void *, const void *);
+} // extern "C"
 
 
 // just for testing...
@@ -476,6 +590,52 @@ int __acc_cdecl_qsort le64_compare_signed(const void *, const void *);
 #endif
 
 
+/*************************************************************************
+// Provide namespaces and classes to abtract endianness policies.
+//
+// CTP - Compile-Time Polymorphism (templates)
+// RTP - Run-Time Polymorphism (virtual functions)
+**************************************************************************/
+
+// forward declarations
+namespace N_BELE_CTP {
+class BEPolicy; class LEPolicy;
+extern const BEPolicy be_policy;
+extern const LEPolicy le_policy;
+}
+namespace N_BELE_RTP {
+class AbstractPolicy;
+class BEPolicy; class LEPolicy;
+extern const BEPolicy be_policy;
+extern const LEPolicy le_policy;
+}
+
+namespace N_BELE_CTP {
+#define BELE_CTP 1
+#include "bele_policy.h"
+#undef BELE_CTP
+}
+
+namespace N_BELE_RTP {
+#define BELE_RTP 1
+#include "bele_policy.h"
+#undef BELE_RTP
+}
+
+namespace N_BELE_CTP {
+
+template <class T>
+inline const N_BELE_RTP::AbstractPolicy* getRTP();
+template <>
+inline const N_BELE_RTP::AbstractPolicy* getRTP<BEPolicy>()
+     { return &N_BELE_RTP::be_policy; }
+template <>
+inline const N_BELE_RTP::AbstractPolicy* getRTP<LEPolicy>()
+     { return &N_BELE_RTP::le_policy; }
+
+}
+
+
 #endif /* already included */
 
 
