@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2004 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2004 Laszlo Molnar
+   Copyright (C) 1996-2007 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2007 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -40,8 +40,16 @@
 #if ((ACC_OS_WIN32 || ACC_OS_WIN64) && ACC_CC_MWERKS) && defined(__MSL__)
 #  undef HAVE_UTIME_H /* this pulls in <windows.h> */
 #endif
+// FIXME - quick hack for arm-wince-gcc-3.4 (Debian pocketpc-*.deb packages)
+#if 1 && (ACC_ARCH_ARM) && defined(__pe__) && !defined(_WIN32)
+#  undef HAVE_CHMOD
+#  undef HAVE_CHOWN
+#  undef HAVE_LSTAT
+#  undef HAVE_UTIME
+#endif
 
 
+// pragmas
 #if (ACC_CC_BORLANDC)
 #  if (__BORLANDC__ < 0x0500)
 #    error "need Borland C++ 5.0 or newer"
@@ -96,6 +104,8 @@
 #  if (_MSC_VER >= 1400)
 #    pragma warning(disable: 4996)      // W1: 'function': was declared deprecated
 #  endif
+#elif (ACC_CC_SUNPROC)
+//#  pragma error_messages(off,"badargtype2w")    // FIXME
 #elif (ACC_CC_WATCOMC)
 #  if (__WATCOMC__ < 1100)
 #    error "need Watcom C++ 11.0c or newer"
@@ -144,35 +154,30 @@
 #  undef __unix
 #endif
 
-
 #if !defined(WITH_UCL)
 #  define WITH_UCL 1
 #endif
-#if defined(WITH_UCL)
-#  include <ucl/uclconf.h>
-#  include <ucl/ucl.h>
-#  if !defined(UCL_VERSION) || (UCL_VERSION < 0x010200L)
-#    error "please upgrade your UCL installation"
-#  endif
-#  if !defined(UPX_UINT_MAX)
-#    define UPX_UINT_MAX  UCL_UINT_MAX
-#    define upx_uint      ucl_uint
-#    define upx_voidp     ucl_voidp
-#    define upx_uintp     ucl_uintp
-#    define upx_byte      ucl_byte
-#    define upx_bytep     ucl_bytep
-#    define upx_bool      ucl_bool
-#    define upx_compress_config_t   ucl_compress_config_t
-#    define upx_progress_callback_t ucl_progress_callback_t
-#    define UPX_E_OK      UCL_E_OK
-#    define UPX_E_ERROR   UCL_E_ERROR
-#    define UPX_E_OUT_OF_MEMORY UCL_E_OUT_OF_MEMORY
-#    define __UPX_CDECL   __UCL_CDECL
+#if 0 && !defined(WITH_LZMA)
+#  define WITH_LZMA 1
+#endif
+#if defined(UPX_OFFICIAL_BUILD)
+#  if !defined(WITH_LZMA) || !defined(WITH_NRV) || !defined(WITH_UCL)
+#    error
 #  endif
 #endif
 #if defined(WITH_NRV)
 #  include <nrv/nrvconf.h>
 #endif
+#if defined(WITH_UCL)
+#  define ucl_compress_config_t REAL_ucl_compress_config_t
+#  include <ucl/uclconf.h>
+#  include <ucl/ucl.h>
+#  if !defined(UCL_VERSION) || (UCL_VERSION < 0x010300L)
+#    error "please upgrade your UCL installation"
+#  endif
+#  undef ucl_compress_config_t
+#  undef ucl_compress_config_p
+#endif
 #if !defined(__UPX_CHECKER)
 #  if defined(__UCL_CHECKER) || defined(__NRV_CHECKER)
 #    define __UPX_CHECKER
@@ -182,6 +187,9 @@
 #  error "UINT_MAX"
 #endif
 
+#define upx_byte                    unsigned char
+#define upx_bytep                   upx_byte *
+
 
 /*************************************************************************
 // system includes
@@ -190,18 +198,6 @@
 // malloc debuggers
 #if defined(WITH_VALGRIND)
 #  include <valgrind/memcheck.h>
-#elif defined(WITH_DMALLOC)
-#  define DMALLOC_FUNC_CHECK
-#  include <dmalloc.h>
-#elif defined(WITH_GC)
-#  define GC_DEBUG
-#  include <gc/gc.h>
-#  undef malloc
-#  undef realloc
-#  undef free
-#  define malloc            GC_MALLOC
-#  define realloc           GC_REALLOC
-#  define free              GC_FREE
 #endif
 
 #if !defined(VALGRIND_MAKE_WRITABLE)
@@ -219,6 +215,7 @@
 #endif
 
 
+// unconditionally turn on assertions
 #undef NDEBUG
 #include <assert.h>
 
@@ -295,18 +292,6 @@
 #undef PAGE_MASK
 #undef PAGE_SIZE
 
-
-#undef __attribute_packed
-#if (ACC_CC_GNUC || ACC_CC_INTELC || ACC_CC_PATHSCALE)
-#  if (1 && (ACC_ARCH_I386))
-#    define __attribute_packed
-#  else
-#    define __attribute_packed      __attribute__((__packed__,__aligned__(1)))
-#  endif
-#else
-#  define __attribute_packed
-#endif
-
 #if !defined(O_BINARY)
 #  define O_BINARY  0
 #endif
@@ -316,6 +301,19 @@
 #endif
 
 
+#undef __attribute_packed
+#if (ACC_CC_GNUC || ACC_CC_INTELC || ACC_CC_PATHSCALE)
+#  if (ACC_ARCH_I386) && (ACC_CC_INTELC && (__INTEL_COMPILER < 800))
+#  elif (0 && (ACC_ARCH_AMD64 || ACC_ARCH_I386))
+#  else
+#    define __attribute_packed      __attribute__((__packed__,__aligned__(1)))
+#  endif
+#endif
+#if !defined(__attribute_packed)
+#  define __attribute_packed
+#endif
+
+
 /*************************************************************************
 //
 **************************************************************************/
@@ -323,68 +321,52 @@
 #define UNUSED(var)              ACC_UNUSED(var)
 #define COMPILE_TIME_ASSERT(e)   ACC_COMPILE_TIME_ASSERT(e)
 
-#if 1
-#  define __COMPILE_TIME_ASSERT_ALIGNOF_SIZEOF(a,b) { \
+#define __COMPILE_TIME_ASSERT_ALIGNOF_SIZEOF(a,b) { \
      typedef a acc_tmp_a_t; typedef b acc_tmp_b_t; \
-     struct acc_tmp_t { acc_tmp_b_t x; acc_tmp_a_t y; acc_tmp_b_t z[7]; } __attribute_packed; \
-     COMPILE_TIME_ASSERT(sizeof(struct acc_tmp_t) == 8*sizeof(b)+sizeof(a)) \
+     struct acc_tmp_t { acc_tmp_b_t x; acc_tmp_a_t y; acc_tmp_b_t z; } __attribute_packed; \
+     COMPILE_TIME_ASSERT(sizeof(struct acc_tmp_t) == 2*sizeof(b)+sizeof(a)) \
    }
-#else
-#  define __COMPILE_TIME_ASSERT_ALIGNOF_SIZEOF(a,b) { \
-     struct acc_tmp_t { b x; a y; b z[7]; } __attribute_packed; \
-     COMPILE_TIME_ASSERT(sizeof(struct acc_tmp_t) == 8*sizeof(b)+sizeof(a)) \
-   }
-#endif
-
 #if defined(__acc_alignof)
-#  define COMPILE_TIME_ASSERT_ALIGNOF(a,b) \
+#  define __COMPILE_TIME_ASSERT_ALIGNOF(a,b) \
      __COMPILE_TIME_ASSERT_ALIGNOF_SIZEOF(a,b) \
      COMPILE_TIME_ASSERT(__acc_alignof(a) == sizeof(b))
 #else
-#  define COMPILE_TIME_ASSERT_ALIGNOF(a,b) \
+#  define __COMPILE_TIME_ASSERT_ALIGNOF(a,b) \
      __COMPILE_TIME_ASSERT_ALIGNOF_SIZEOF(a,b)
 #endif
+#define COMPILE_TIME_ASSERT_ALIGNED1(a)     __COMPILE_TIME_ASSERT_ALIGNOF(a,char)
 
 #define TABLESIZE(table)    ((sizeof(table)/sizeof((table)[0])))
 
+
+#if 0
 #define ALIGN_DOWN(a,b)     (((a) / (b)) * (b))
 #define ALIGN_UP(a,b)       ALIGN_DOWN((a) + ((b) - 1), b)
 #define ALIGN_GAP(a,b)      (ALIGN_UP(a,b) - (a))
+#elif 1
+template <class T>
+inline T ALIGN_DOWN(const T& a, const T& b) { T r; r = (a / b) * b; return r; }
+template <class T>
+inline T ALIGN_UP  (const T& a, const T& b) { T r; r = ((a + b - 1) / b) * b; return r; }
+template <class T>
+inline T ALIGN_GAP (const T& a, const T& b) { T r; r = ALIGN_UP(a, b) - a; return r; }
+#else
+inline unsigned ALIGN_DOWN(unsigned a, unsigned b) { return (a / b) * b; }
+inline unsigned ALIGN_UP  (unsigned a, unsigned b) { return ((a + b - 1) / b) * b; }
+inline unsigned ALIGN_GAP (unsigned a, unsigned b) { return ALIGN_UP(a, b) - a; }
+#endif
 
+#if 0
 #define UPX_MAX(a,b)        ((a) >= (b) ? (a) : (b))
 #define UPX_MIN(a,b)        ((a) <= (b) ? (a) : (b))
-#define UPX_MAX3(a,b,c)     ((a) >= (b) ? UPX_MAX(a,c) : UPX_MAX(b,c))
-#define UPX_MIN3(a,b,c)     ((a) <= (b) ? UPX_MIN(a,c) : UPX_MIN(b,c))
-
-
-#if 0 && defined(__cplusplus) && !defined(new) && !defined(delete)
-// global operators - debug
-inline void *operator new(size_t l)
-{
-    void *p = malloc(l);
-    printf("new   %6ld %p\n",(long)l,p);
-    fflush(stdout);
-    return p;
-}
-inline void *operator new[](size_t l)
-{
-    void *p = malloc(l);
-    printf("new[] %6ld %p\n",(long)l,p);
-    fflush(stdout);
-    return p;
-}
-inline void operator delete(void *p)
-{
-    printf("delete       %p\n",p);
-    fflush(stdout);
-    if (p) free(p);
-}
-inline void operator delete[](void *p)
-{
-    printf("delete[]     %p\n",p);
-    fflush(stdout);
-    if (p) free(p);
-}
+#elif 1
+template <class T>
+inline const T& UPX_MAX(const T& a, const T& b) { if (a < b) return b; return a; }
+template <class T>
+inline const T& UPX_MIN(const T& a, const T& b) { if (a < b) return a; return b; }
+#else
+inline unsigned UPX_MAX(unsigned a, unsigned b) { return a < b ? b : a; }
+inline unsigned UPX_MIN(unsigned a, unsigned b) { return a < b ? a : b; }
 #endif
 
 
@@ -397,6 +379,14 @@ inline void operator delete[](void *p)
 
 #define ByteArray(var, size)    Array(unsigned char, var, size)
 
+struct nocopy
+{
+    inline nocopy() {}
+private:
+    nocopy(const nocopy &); // undefined
+    nocopy& operator=(const nocopy &); // undefined
+};
+
 
 /*************************************************************************
 // constants
@@ -416,24 +406,23 @@ inline void operator delete[](void *p)
 #define EXIT_INTERNAL   1
 
 
-// compression methods - DO NOT CHANGE
-#define M_NRV2B_LE32    2
-#define M_NRV2B_8       3
-#define M_NRV2B_LE16    4
-#define M_NRV2D_LE32    5
-#define M_NRV2D_8       6
-#define M_NRV2D_LE16    7
-#define M_NRV2E_LE32    8
-#define M_NRV2E_8       9
-#define M_NRV2E_LE16    10
-#define M_CL1B_LE32     11
-#define M_CL1B_8        12
-#define M_CL1B_LE16     13
+// magic constants for patching
+#define UPX_MAGIC_LE32          0x21585055      /* "UPX!" */
+#define UPX_MAGIC2_LE32         0xD5D0D8A1
 
-#define M_IS_NRV2B(x)   ((x) >= M_NRV2B_LE32 && (x) <= M_NRV2B_LE16)
-#define M_IS_NRV2D(x)   ((x) >= M_NRV2D_LE32 && (x) <= M_NRV2D_LE16)
-#define M_IS_NRV2E(x)   ((x) >= M_NRV2E_LE32 && (x) <= M_NRV2E_LE16)
-#define M_IS_CL1B(x)    ((x) >= M_CL1B_LE32  && (x) <= M_CL1B_LE16)
+
+// upx_compress() error codes
+#define UPX_E_OK                    (0)
+#define UPX_E_ERROR                 (-1)
+#define UPX_E_OUT_OF_MEMORY         (-2)
+#define UPX_E_NOT_COMPRESSIBLE      (-3)
+#define UPX_E_INPUT_OVERRUN         (-4)
+#define UPX_E_OUTPUT_OVERRUN        (-5)
+#define UPX_E_LOOKBEHIND_OVERRUN    (-6)
+#define UPX_E_EOF_NOT_FOUND         (-7)
+#define UPX_E_INPUT_NOT_CONSUMED    (-8)
+#define UPX_E_NOT_YET_IMPLEMENTED   (-9)
+#define UPX_E_INVALID_ARGUMENT      (-10)
 
 
 // Executable formats. Note: big endian types are >= 128.
@@ -459,6 +448,14 @@ inline void operator delete[](void *p)
 #define UPX_F_LINUX_ELFI_i386   20
 #define UPX_F_WINCE_ARM_PE      21
 #define UPX_F_LINUX_ELF64_AMD   22
+#define UPX_F_LINUX_ELF32_ARMEL 23
+#define UPX_F_BSD_i386          24
+#define UPX_F_BSD_ELF_i386      25
+#define UPX_F_BSD_SH_i386       26
+
+#define UPX_F_VMLINUX_AMD64     27
+#define UPX_F_VMLINUX_ARMEL     28
+#define UPX_F_MACH_i386         29
 
 #define UPX_F_PLAIN_TEXT        127
 
@@ -466,10 +463,231 @@ inline void operator delete[](void *p)
 #define UPX_F_SOLARIS_SPARC     130
 #define UPX_F_MACH_PPC32        131
 #define UPX_F_LINUX_ELFPPC32    132
+#define UPX_F_LINUX_ELF32_ARMEB 133
+#define UPX_F_MACH_FAT          134
+#define UPX_F_VMLINUX_ARMEB     135
 
 
-#define UPX_MAGIC_LE32      0x21585055          /* "UPX!" */
-#define UPX_MAGIC2_LE32     0xD5D0D8A1
+// compression methods
+#define M_ALL           (-1)
+#define M_END           (-2)
+#define M_NONE          (-3)
+#define M_SKIP          (-4)
+#define M_ULTRA_BRUTE   (-5)
+// compression methods - DO NOT CHANGE
+#define M_NRV2B_LE32    2
+#define M_NRV2B_8       3
+#define M_NRV2B_LE16    4
+#define M_NRV2D_LE32    5
+#define M_NRV2D_8       6
+#define M_NRV2D_LE16    7
+#define M_NRV2E_LE32    8
+#define M_NRV2E_8       9
+#define M_NRV2E_LE16    10
+//#define M_CL1B_LE32     11
+//#define M_CL1B_8        12
+//#define M_CL1B_LE16     13
+#define M_LZMA          14
+#define M_DEFLATE       15      /* zlib */
+
+#define M_IS_NRV2B(x)   ((x) >= M_NRV2B_LE32 && (x) <= M_NRV2B_LE16)
+#define M_IS_NRV2D(x)   ((x) >= M_NRV2D_LE32 && (x) <= M_NRV2D_LE16)
+#define M_IS_NRV2E(x)   ((x) >= M_NRV2E_LE32 && (x) <= M_NRV2E_LE16)
+//#define M_IS_CL1B(x)    ((x) >= M_CL1B_LE32  && (x) <= M_CL1B_LE16)
+#define M_IS_LZMA(x)    (((x) & 255) == M_LZMA)
+#define M_IS_DEFLATE(x) ((x) == M_DEFLATE)
+
+
+// filters
+#define FT_END          (-1)
+#define FT_NONE         (-2)
+#define FT_SKIP         (-3)
+#define FT_ULTRA_BRUTE  (-4)
+
+
+/*************************************************************************
+// compression - callback_t
+**************************************************************************/
+
+struct upx_callback_t;
+#define upx_callback_p upx_callback_t *
+#if 0
+typedef void* (__acc_cdecl *upx_alloc_func_t)
+    (upx_callback_p self, unsigned items, unsigned size);
+typedef void      (__acc_cdecl *upx_free_func_t)
+    (upx_callback_p self, void* ptr);
+#endif
+typedef void (__acc_cdecl *upx_progress_func_t)
+    (upx_callback_p, unsigned, unsigned);
+
+struct upx_callback_t
+{
+    upx_progress_func_t nprogress;
+    void *user;
+
+    void reset() { memset(this, 0, sizeof(*this)); }
+};
+
+
+/*************************************************************************
+// compression - config_t
+**************************************************************************/
+
+template <class T, T default_value, T min_value, T max_value>
+struct OptVar
+{
+    typedef T Type;
+    static const T default_value_c = default_value;
+    static const T min_value_c = min_value;
+    static const T max_value_c = max_value;
+
+    void assertValue() {
+        // FIXME: this generates annoying warnings "unsigned >= 0 is always true"
+        //assert((v >= min_value) && (v <= max_value));
+    }
+
+    OptVar() : v(default_value), is_set(0) { }
+    OptVar& operator= (const T other) {
+        v = other; is_set += 1;
+        assertValue();
+        return *this;
+    }
+#if 0
+#error
+    // there is too much implicit magic in this copy operator;
+    // better introduce an explicit "oassign" function.
+    OptVar& operator= (const OptVar& other) {
+        if (other.is_set) { v = other.v; is_set += 1; }
+        assertValue(); return *this;
+    }
+#endif
+
+    void reset() { v = default_value; is_set = 0; }
+    operator T () const { return v; }
+
+    T v;
+    unsigned is_set;
+};
+
+
+// optional assignments
+template <class T> inline void oassign(T& self, const T& other) {
+    if (other.is_set) { self.v = other.v; self.is_set += 1; }
+}
+#if 0
+template <class V, class T> inline void oassign(V& v, const T& other) {
+    if (other.is_set) { v = other.v; }
+}
+#else
+template <class T> inline void oassign(unsigned& v, const T& other) {
+    if (other.is_set) { v = other.v; }
+}
+#endif
+
+
+struct lzma_compress_config_t
+{
+    typedef OptVar<unsigned,  2u, 0u,   4u> pos_bits_t;             // pb
+    typedef OptVar<unsigned,  0u, 0u,   4u> lit_pos_bits_t;         // lb
+    typedef OptVar<unsigned,  3u, 0u,   8u> lit_context_bits_t;     // lc
+    typedef OptVar<unsigned, (1u<<22), 1u, (1u<<30) > dict_size_t;
+    typedef OptVar<unsigned, 64u, 5u, 273u> num_fast_bytes_t;
+
+    pos_bits_t          pos_bits;           // pb
+    lit_pos_bits_t      lit_pos_bits;       // lp
+    lit_context_bits_t  lit_context_bits;   // lc
+    dict_size_t         dict_size;
+    unsigned            fast_mode;
+    num_fast_bytes_t    num_fast_bytes;
+    unsigned            match_finder_cycles;
+
+    unsigned            max_num_probs;
+
+    void reset();
+};
+
+
+struct ucl_compress_config_t : public REAL_ucl_compress_config_t
+{
+    void reset() { memset(this, 0xff, sizeof(*this)); }
+};
+
+
+struct zlib_compress_config_t
+{
+    typedef OptVar<unsigned,  8u, 1u,   9u> mem_level_t;            // ml
+    typedef OptVar<unsigned, 15u, 9u,  15u> window_bits_t;          // wb
+    typedef OptVar<unsigned,  0u, 0u,   4u> strategy_t;             // st
+
+    mem_level_t         mem_level;          // ml
+    window_bits_t       window_bits;        // wb
+    strategy_t          strategy;           // st
+
+    void reset();
+};
+
+
+struct upx_compress_config_t
+{
+    lzma_compress_config_t  conf_lzma;
+    ucl_compress_config_t   conf_ucl;
+    zlib_compress_config_t  conf_zlib;
+    void reset() { conf_lzma.reset(); conf_ucl.reset(); conf_zlib.reset(); }
+};
+
+#define NULL_cconf  ((upx_compress_config_t *) NULL)
+
+
+/*************************************************************************
+// compression - result_t
+**************************************************************************/
+
+struct lzma_compress_result_t
+{
+    unsigned pos_bits;              // pb
+    unsigned lit_pos_bits;          // lp
+    unsigned lit_context_bits;      // lc
+    unsigned dict_size;
+    unsigned fast_mode;
+    unsigned num_fast_bytes;
+    unsigned match_finder_cycles;
+    unsigned num_probs;             // (computed result)
+
+    void reset() { memset(this, 0, sizeof(*this)); }
+};
+
+
+struct ucl_compress_result_t
+{
+    ucl_uint result[16];
+
+    void reset() { memset(this, 0, sizeof(*this)); }
+};
+
+
+struct zlib_compress_result_t
+{
+    unsigned dummy;
+
+    void reset() { memset(this, 0, sizeof(*this)); }
+};
+
+
+struct upx_compress_result_t
+{
+    // debug
+    int method, level;
+    unsigned u_len, c_len;
+
+    lzma_compress_result_t  result_lzma;
+    ucl_compress_result_t   result_ucl;
+    zlib_compress_result_t  result_zlib;
+
+    void reset() {
+        memset(this, 0, sizeof(*this));
+        result_lzma.reset(); result_ucl.reset(); result_zlib.reset();
+    }
+};
 
 
 /*************************************************************************
@@ -488,9 +706,13 @@ inline void operator delete[](void *p)
 #include "console.h"
 
 
+// classes
+class ElfLinker;
+typedef ElfLinker Linker;
+
+
 // main.cpp
 extern const char *progname;
-void init_options(struct options_t *o);
 bool set_ec(int ec);
 #if (ACC_CC_GNUC || ACC_CC_LLVM || ACC_CC_PATHSCALE)
 void e_exit(int ec) __attribute__((__noreturn__));
@@ -537,7 +759,7 @@ void do_files(int i, int argc, char *argv[]);
 
 // help.cpp
 void show_head(void);
-void show_help(int x = 0);
+void show_help(int verbose=0);
 void show_license(void);
 void show_usage(void);
 void show_version(int);
@@ -547,18 +769,22 @@ void show_version(int);
 unsigned upx_adler32(const void *buf, unsigned len, unsigned adler=1);
 unsigned upx_crc32(const void *buf, unsigned len, unsigned crc=0);
 
-int upx_compress           ( const upx_bytep src, upx_uint  src_len,
-                                   upx_bytep dst, upx_uintp dst_len,
-                                   upx_progress_callback_t *cb,
+int upx_compress           ( const upx_bytep src, unsigned  src_len,
+                                   upx_bytep dst, unsigned* dst_len,
+                                   upx_callback_p cb,
                                    int method, int level,
-                             const struct upx_compress_config_t *conf,
-                                   upx_uintp result );
-int upx_decompress         ( const upx_bytep src, upx_uint  src_len,
-                                   upx_bytep dst, upx_uintp dst_len,
-                                   int method );
-int upx_test_overlap       ( const upx_bytep buf, upx_uint src_off,
-                                   upx_uint  src_len, upx_uintp dst_len,
-                                   int method );
+                             const upx_compress_config_t *cconf,
+                                   upx_compress_result_t *cresult );
+int upx_decompress         ( const upx_bytep src, unsigned  src_len,
+                                   upx_bytep dst, unsigned* dst_len,
+                                   int method,
+                             const upx_compress_result_t *cresult );
+int upx_test_overlap       ( const upx_bytep buf,
+                             const upx_bytep tbuf,
+                                   unsigned  src_off, unsigned src_len,
+                                   unsigned* dst_len,
+                                   int method,
+                             const upx_compress_result_t *cresult );
 
 
 #endif /* __cplusplus */
