@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2004 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2004 Laszlo Molnar
+   Copyright (C) 1996-2007 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2007 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -47,7 +47,7 @@ enum {
 };
 
 
-struct UiPacker__State
+struct UiPacker::State
 {
     int mode;
 
@@ -92,8 +92,6 @@ long UiPacker::update_u_len = 0;
 long UiPacker::update_fc_len = 0;
 long UiPacker::update_fu_len = 0;
 
-#define clear_cb()    memset(&cb, 0, sizeof(cb))
-
 
 /*************************************************************************
 // constants
@@ -178,13 +176,13 @@ static const char *mkline(unsigned long fu_len, unsigned long fc_len,
 **************************************************************************/
 
 UiPacker::UiPacker(const Packer *p_) :
-    p(p_), s(NULL)
+    ui_pass(0), ui_total_passes(0), p(p_), s(NULL)
 {
     init_global_constants();
 
-    clear_cb();
+    cb.reset();
 
-    s = new UiPacker__State;
+    s = new State;
     memset(s,0,sizeof(*s));
     s->msg_buf[0] = '\r';
 
@@ -208,7 +206,7 @@ UiPacker::UiPacker(const Packer *p_) :
 
 UiPacker::~UiPacker()
 {
-    clear_cb();
+    cb.reset();
     delete s; s = NULL;
 }
 
@@ -243,10 +241,10 @@ void UiPacker::startCallback(unsigned u_len, unsigned step,
     s->bar_len = 64;
     s->pos = -2;
     s->spin_counter = 0;
-    s->bar_pos = 1;             // because of the leading `\r'
+    s->bar_pos = 1;             // because of the leading '\r'
     s->pass_digits = 0;
 
-    clear_cb();
+    cb.reset();
 
     if (s->pass < 0)            // no callback wanted
         return;
@@ -265,11 +263,11 @@ void UiPacker::startCallback(unsigned u_len, unsigned step,
     }
 
 #if (ACC_CC_MSC && (_MSC_VER == 1300))
-    cb.callback = &UiPacker::callback;
+    cb.nprogress = &UiPacker::progress_callback;
 #else
-    cb.callback = callback;
+    cb.nprogress = progress_callback;
 #endif
-    cb.user = this;
+    cb.user = this; // parameter for static function UiPacker::progress_callback()
 
     if (s->mode == M_CB_TERM)
     {
@@ -386,7 +384,7 @@ void UiPacker::endCallback()
     }
 #endif /* UI_USE_SCREEN */
 
-    clear_cb();
+    cb.reset();
 #if 0
     printf("callback: pass %d, step %6d, updates %6d\n",
            s->pass, s->step, s->spin_counter);
@@ -398,15 +396,11 @@ void UiPacker::endCallback()
 // the callback
 **************************************************************************/
 
-void __UPX_CDECL UiPacker::callback(upx_uint isize, upx_uint osize, int state, void *user)
+void __acc_cdecl UiPacker::progress_callback(upx_callback_p cb, unsigned isize, unsigned osize)
 {
     //printf("%6d %6d %d\n", isize, osize, state);
-    if (state != -1 && state != 3) return;
-    if (user)
-    {
-        UiPacker *uip = (UiPacker *) user;
-        uip->doCallback(isize, osize);
-    }
+    UiPacker *self = (UiPacker *) cb->user;
+    self->doCallback(isize, osize);
 }
 
 
@@ -463,7 +457,7 @@ void UiPacker::doCallback(unsigned isize, unsigned osize)
     upx_snprintf(m, buflen, "  %3d.%1d%%  %c ",
                  ratio / 10000, (ratio % 10000) / 1000,
                  spinner[s->spin_counter & 3]);
-    assert((int)strlen(s->msg_buf) < 1 + 80);
+    assert(strlen(s->msg_buf) < 1 + 80);
 
     s->pos = pos;
     s->spin_counter++;
@@ -487,7 +481,7 @@ void UiPacker::doCallback(unsigned isize, unsigned osize)
 #if 0
         s->screen->putString(s->screen,msg,s->b_cx,s->b_cy);
 #else
-        // FIXME: this doesn't honor `--mono' etc.
+        // FIXME: this doesn't honor '--mono' etc.
         int attr = FG_CYAN | s->s_bg;
         s->screen->putStringAttr(s->screen,msg,attr,s->b_cx,s->b_cy);
 #endif
@@ -510,7 +504,7 @@ void UiPacker::uiPackStart(const OutputFile *fo)
 
 void UiPacker::uiPackEnd(const OutputFile *fo)
 {
-    uiUpdate(fo->getBytesWritten());
+    uiUpdate(fo->st_size());
 
     if (s->mode == M_QUIET)
         return;
@@ -527,7 +521,7 @@ void UiPacker::uiPackEnd(const OutputFile *fo)
     else if (opt->to_stdout)
         name = "<stdout>";
     con_fprintf(stdout,"%s\n",
-                mkline(p->ph.u_file_size, fo->getBytesWritten(),
+                mkline(p->ph.u_file_size, fo->st_size(),
                        p->ph.u_len, p->ph.c_len,
                        p->getName(), fn_basename(name)));
     printSetNl(0);
@@ -669,7 +663,7 @@ bool UiPacker::uiFileInfoStart()
     total_files++;
 
     int fg = con_fg(stdout,FG_CYAN);
-    con_fprintf(stdout,"%s [%s]\n", p->fi->getName(), p->getName());
+    con_fprintf(stdout,"%s [%s, %s]\n", p->fi->getName(), p->getFullName(opt), p->getName());
     fg = con_fg(stdout,fg);
     UNUSED(fg);
     if (p->ph.c_len > 0)
